/***************************************************************************
 *   Copyright (C) 2005 by David Voswinkel                                 *
 *   d.voswinkel@netcologne.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "trayicon.h"
#include "pref.h"
#include "osd.h"
#include "kbtscoworker.h"

#include <algorithm>
#include <kaction.h>
#include <kapplication.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kglobal.h>
#include <khelpmenu.h>
#include <kiconloader.h>
#include <kicontheme.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <knotifydialog.h>
#include <knotifyclient.h>
#include <kpixmapeffect.h>
#include <kpopupmenu.h>
#include <kprocess.h>
#include <krun.h>
#include <qevent.h>
#include <qtooltip.h>

TrayIcon::TrayIcon(KBtsco* app)
        : state(STATE_IDLE),
          state_last(STATE_NONE)

{
    iconConnectingBlinkState = false;
    iconConnectedSCOBlinkState = false;
    iconConnectedBlinkState = 0;
    this->helpMenu = new KHelpMenu(this, KApplication::kApplication()->aboutData());
    this->app = app;


    setCaption("KBtsco");

    KConfig *config = KGlobal::config();
    config->setGroup("UI");

    m_config = new KConfig("kbtscorc");

    blinkTimer = new QTimer(this);
    connect(blinkTimer, SIGNAL(timeout()), this, SLOT(updateIcon()));

    KPixmap appPixmap = KGlobal::iconLoader()->loadIcon(
            "kbtsco", KIcon::Small, 16);

    contextMenu()->changeTitle( 0, appPixmap, "KBtsco");

    // Connect entry
    headsetConnectAction = new KAction(i18n("&Connect"),
                                    KGlobal::iconLoader()->loadIcon("connect_creating", KIcon::Small, 16),
                                    KShortcut::null(), this, "headset_connect");
    connect(headsetConnectAction, SIGNAL(activated()), this, SLOT(slotHeadsetConnect()));
    headsetConnectAction->plug(contextMenu());

    // Disconnect entry
    headsetDisconnectAction = new KAction(i18n("&Disconnect"),
                                    KGlobal::iconLoader()->loadIcon("connect_no", KIcon::Small, 16),
                                    KShortcut::null(), this, "headset_disconnect");
    connect(headsetDisconnectAction, SIGNAL(activated()), this, SLOT(slotHeadsetDisconnect()));
    headsetDisconnectAction->plug(contextMenu());
    headsetDisconnectAction->setEnabled(false);


    // Configuration menu
    configActionMenu = new KActionMenu(i18n("&Configuration"),
        KGlobal::iconLoader()->loadIcon("configure", KIcon::Small, 16), this, "config_menu");
    configActionMenu->plug(contextMenu());

    // Help menu
    helpActionMenu = new KActionMenu(i18n("&Help"),
        KGlobal::iconLoader()->loadIcon("help", KIcon::Small, 16), this, "help_menu");
    helpActionMenu->plug(contextMenu());





    // Menu entries:
    // -------------

    // Service configuration
    headsetConfigAction = new KAction(i18n("Configure &Headset"),
        KGlobal::iconLoader()->loadIcon("configure", KIcon::Small, 16),
        KShortcut::null(), this, "headset_config");
    connect(headsetConfigAction, SIGNAL(activated()), this, SLOT(slotHeadsetConfig()));
    configActionMenu->insert(headsetConfigAction);



    // show documentation
    KAction* showHelpAction = KStdAction::help(this,
        SLOT(slotShowHelp()), actionCollection());
    helpActionMenu->insert(showHelpAction);

    // Report bug menu item
    KAction* reportBugAction = KStdAction::reportBug(this,
        SLOT(slotReportBug()), actionCollection());
    helpActionMenu->insert(reportBugAction);

    // "About" menu item
    KAction* aboutAction = KStdAction::aboutApp(this,
        SLOT(slotAbout()), actionCollection());
    helpActionMenu->insert(aboutAction);


    KPixmap logoPixmap = KGlobal::iconLoader()->loadIcon(
        "kbtsco", KIcon::Small, 22);

    KPixmap logoConnectedPixmap = KGlobal::iconLoader()->loadIcon(
            "kbtsco_active", KIcon::Small, 22);

    KPixmap logoActivePixmap = KGlobal::iconLoader()->loadIcon(
            "kbtsco_active", KIcon::Small, 22);


    iconIdle            = logoPixmap;
    iconConnecting      = logoPixmap;
    iconConnected       = logoConnectedPixmap;
    iconConnectedsco    = logoActivePixmap;
    iconNoAddress = logoPixmap;
    KPixmapEffect::toGray(iconIdle);
    KPixmapEffect::fade(iconIdle, 0.3, QColor(255,255,255));
    KPixmapEffect::toGray(iconNoAddress);
    KPixmapEffect::fade(iconNoAddress, 0.5, QColor(128,128,128));



    m_config->setGroup ("bluetooth");
    m_dev_baddr = m_config->readEntry("baddr");
    m_dev_name  = m_config->readEntry("name");

    m_worker = new KBtscoWorker(this);
    m_worker->setBaddr (m_dev_baddr);
    m_worker->start();

    m_osd = new OSDWidget ("KBtsco", this, "KBtsco");
    show();
    setConnectionState(STATE_IDLE);
    connect(this, SIGNAL(quitSelected()), this, SLOT(slotQuitSelected()));


}

TrayIcon::~TrayIcon()
{
    kdDebug (200011) << "TrayIcon::~TrayIcon" << endl;
    if (m_worker){
        kdDebug (200011) << "TrayIcon::~TrayIcon m_worker seem allocated, will delete" << endl;
        delete m_worker;
        m_worker = 0;
    }
}

void TrayIcon::customEvent (QCustomEvent * e)
{

    KBtscoStatusEvent::Type t = (KBtscoStatusEvent::Type) e->type ();
    KBtscoStatusEvent *se = (KBtscoStatusEvent *) e;
    kdDebug (200011) << "TrayIcon::customEvent e=" << e << " t=" << t  << endl;
    switch (t) {
        case KBtscoStatusEvent::Message:
            m_osd->showOSD (se->string (), TRUE);
            break;
        case KBtscoStatusEvent::Error:
            KMessageBox::error (this, se->string ());
            break;
        case KBtscoStatusEvent::Volume:
            break;
        case KBtscoStatusEvent::Connecting:
            setConnectionState(STATE_IDLE);
            break;
        case KBtscoStatusEvent::Connected:
            setConnectionState( STATE_CONNECTED);
            break;
        case KBtscoStatusEvent::ConnectedSCO:
            setConnectionState( STATE_CONNECTED_SCO);
            break;
        case KBtscoStatusEvent::DisconnectedSCO:
            setConnectionState(STATE_CONNECTED);
            break;
        case KBtscoStatusEvent::Disconnected:
            setConnectionState(STATE_IDLE);
            break;
        default:
            break;

    }
}

void TrayIcon::updateIcon()
{
    QString oldText = QToolTip::textFor(this);
    QString newText = "";
    if ( STATE_IDLE  == state ) {
        if (blinkTimer->isActive())
            blinkTimer->stop();
        if (state_last != state){
            if (hasAddress){
                setPixmap(iconIdle);
                newText = i18n("Not connected");
            } else {
                setPixmap(iconNoAddress);
                newText = i18n("No Device configured");
            }
        }
    } else if (STATE_CONNECTING ==  state ) {
        if (state_last != state){
            if (blinkTimer->isActive())
                blinkTimer->stop();
            blinkTimer->start(250);
            newText = i18n("Connecting to <b>") + m_dev_name + " [" + m_dev_baddr + "]</b>";
        }
        if (!blinkTimer->isActive())
            blinkTimer->start(250);
        iconConnectingBlinkState = !iconConnectingBlinkState;
        if (iconConnectingBlinkState) {
            setPixmap(iconConnecting);
        } else {
            setPixmap(iconIdle);
        }
    } else if (STATE_CONNECTED == state ) {
        if (state_last != state){
            if (blinkTimer->isActive())
                blinkTimer->stop();
            blinkTimer->start(500);
            newText = i18n("Connected to <b>") + m_dev_name + " [" + m_dev_baddr + "]</b>";
            iconConnectedBlinkState = 0;
        }
        if (!blinkTimer->isActive())
            blinkTimer->start(500);
        iconConnectedBlinkState++;
        if (!(iconConnectedBlinkState%4)) {
            iconConnectedBlinkState = 0;
            setPixmap(iconConnected);
        } else {
            setPixmap(iconConnecting);
        }
    }else if (STATE_CONNECTED_SCO == state ) {
        if (state_last != state){
            if (blinkTimer->isActive())
                blinkTimer->stop();
            blinkTimer->start(250);
            newText = i18n("SCO active to <b>") + m_dev_name + " [" + m_dev_baddr + "]</b>";
        }
        if (!blinkTimer->isActive())
            blinkTimer->start(250);
        iconConnectedSCOBlinkState = !iconConnectedSCOBlinkState;
        if (iconConnectedSCOBlinkState) {
            setPixmap(iconConnectedsco);
        } else {
            setPixmap(iconConnecting);
        }
    }
    if (state_last != state && oldText != newText) {
        QToolTip::remove(this);
        QToolTip::add(this, newText);
    }
    state_last = state;
}


void TrayIcon::mousePressEvent(QMouseEvent *e)
{
    if (e->button() == QMouseEvent::LeftButton) {
        e->accept();
        if (STATE_IDLE == state){
            slotHeadsetConnect();
        } else {
            slotHeadsetDisconnect();
        }
    }
    else {
        KSystemTray::mousePressEvent(e);
    }
}

void TrayIcon::closeEvent(QCloseEvent *e)
{
    if (acceptClose) {
        QWidget::closeEvent(e);
    }
    else {
        e->ignore();
    }
    acceptClose = true;
}


void TrayIcon::enterEvent (QEvent *e )
{
    kdDebug (200011) << "TrayIcon::enterEvent" << endl;
}


void TrayIcon::slotQuitSelected()
{
    // Ask if the user want to simply quit or disable
    // automatic start of kbluetoothd

    if (m_worker){
        delete m_worker;
        m_worker = 0;
    }

    int autoStart = KMessageBox::questionYesNoCancel( 0,
        i18n("Should KBtsco still be restarted when you login?"),
        i18n("Automatically start KBtsco?") );
    acceptClose = true;
    KConfig *config = KGlobal::config();
    config->setGroup("General");
    if (autoStart == KMessageBox::Yes) {
        config->writeEntry("AutoStart", true);
    }
    else if (autoStart == KMessageBox::No) {
        config->writeEntry("AutoStart", false);
    } else {
        acceptClose = false;
        return;
    }
    KMessageBox::information(this,
        i18n("You can reenable the KBtsco daemon at any time by running 'kbtsco'."),
        i18n("KDE Btsco daemon"), "quitrestart_messagebox");

    config->sync();
}

void TrayIcon::slotReportBug()
{
    helpMenu->reportBug();
}

void TrayIcon::slotAbout()
{
    helpMenu->aboutApplication();
}

void TrayIcon::slotShowHelp()
{
    KApplication::kApplication()->invokeHelp("","kbtsco");
}

void TrayIcon::slotHeadsetConfig()
{
    kdDebug (200011) << "TrayIcon::slotHeadsetConfig"<< endl;
    KBtscoPreferences dlg (m_config);
    if (dlg.exec ()) {
        m_config->setGroup ("bluetooth");
        m_dev_baddr = m_config->readEntry("baddr");
        m_dev_name  = m_config->readEntry("name");
        hasAddress = !m_dev_baddr.isEmpty();
        kdDebug (200011) << "TrayIcon::slotHeadsetConfig m_dev_baddr=" << m_dev_baddr<< endl;
        kdDebug (200011) << "TrayIcon::slotHeadsetConfig m_dev_name=" << m_dev_name << endl;
        kdDebug (200011) << "TrayIcon::slotHeadsetConfig hasAddress="<< hasAddress << endl;
        m_worker->setBaddr(m_dev_baddr);
        setConnectionState(STATE_NONE);
    }
}

void TrayIcon::slotHeadsetConnect()
{
    kdDebug (200011) << "TrayIcon::slotHeadsetConnect"<< endl;
    if (!hasAddress){
        KMessageBox::error(0,i18n("Please set the address of the bluetooth device first."),i18n("Connect Error"));
        slotHeadsetConfig();
        return;
    }
    setConnectionState(STATE_CONNECTING);
    KBtscoWorkerCommand *c = new KBtscoWorkerCommand (KBtscoWorkerCommand::Connect, 0);
    m_worker->addCommand (c);
}

void TrayIcon::slotHeadsetDisconnect()
{
    kdDebug (200011) << "TrayIcon::slotHeadsetDisconnect"<< endl;

    if (STATE_IDLE == state)
        return;
    KBtscoWorkerCommand *c = new KBtscoWorkerCommand (KBtscoWorkerCommand::Disconnect, 0);
    m_worker->addCommand (c);
}


void TrayIcon::setConnectionState(state_t s)
{
    QString osd_message;

    hasAddress = !m_dev_baddr.isEmpty();
    m_config->setGroup("application");
    osdEnabled = ("true" == m_config->readEntry("osd_enable"));
    if (STATE_NONE != s)
        state = s;

    switch(state){
        case STATE_NONE:
        case STATE_IDLE:
            kdDebug (200011) << "TrayIcon::setConnectionState STATE_IDLE"<< endl;
            headsetDisconnectAction->setEnabled(false);
            headsetConnectAction->setEnabled(true);
            if (osdEnabled){
                if (hasAddress){
                    if (!m_dev_name.isEmpty())
                        osd_message = i18n("Idle Headset ") + m_dev_name + " [" + m_dev_baddr + "]";
                    else
                        osd_message = i18n("Idle Headset ") + m_dev_baddr;
                } else {
                    osd_message = i18n("No Device") + m_dev_baddr;
                }
                m_osd->showOSD(osd_message, TRUE);
            }
            updateIcon();
            break;
        case STATE_CONNECTING:
            kdDebug (200011) << "TrayIcon::setConnectionState STATE_CONNECTING"<< endl;
            headsetConnectAction->setEnabled(false);
            headsetDisconnectAction->setEnabled(false);
            updateIcon();

            if (osdEnabled){
                if (!m_dev_name.isEmpty())
                    osd_message = i18n("Connecting to ") + m_dev_name + " [" + m_dev_baddr + "]";
                else
                    osd_message = i18n("Connecting to ") + m_dev_baddr;
                m_osd->showOSD(osd_message, TRUE);
            }
            break;
        case STATE_CONNECTED:
            kdDebug (200011) << "TrayIcon::setConnectionState STATE_CONNECTED"<< endl;
            headsetConnectAction->setEnabled(false);
            headsetDisconnectAction->setEnabled(true);
            if (STATE_CONNECTED_SCO == state_last){
                if (osdEnabled){
                    if (!m_dev_name.isEmpty())
                        osd_message = i18n("Release SCO ") + m_dev_name + " [" + m_dev_baddr + "]";
                    else
                        osd_message = i18n("Release SCO ") + m_dev_baddr;
                    m_osd->showOSD (osd_message, TRUE);
                }
            } else {
                if (osdEnabled){
                    if (!m_dev_name.isEmpty())
                        osd_message = i18n("Connected to ") + m_dev_name + " [" + m_dev_baddr + "]";
                    else
                        osd_message = i18n("Connected to ") + m_dev_baddr;
                    m_osd->showOSD (osd_message, TRUE);
                }
            }
            updateIcon();
            break;
        case STATE_CONNECTED_SCO:
            kdDebug (200011) << "TrayIcon::setConnectionState STATE_CONNECTED_SCO"<< endl;
            if (osdEnabled){
                if (!m_dev_name.isEmpty())
                    osd_message = i18n("SCO active ") + m_dev_name + " [" + m_dev_baddr + "]";
                else
                    osd_message = i18n("SCO active ") + m_dev_baddr;
                m_osd->showOSD (osd_message, TRUE);
            }
            updateIcon();
            break;
    }
}


#include "trayicon.moc"
